#include "header\gba.h"
#include "header\bg.h"
#include "header\timer.h"
#include "header\dma.h"
#include "header\irq.h"
#include "header\sound.h"
#include "header\sprite.h"
#include "header\blend.h"

#include "header\font\BGFont.h"
//#include "header\font\BitMapFont.h"

#include "lib.h"
#include "syscall.h"
#include "resource.h"

#include <stdlib.h>

//Mode0
//BG2 ^C
//BG3 \

//---------------------------------------------------------------------------
//
#define BG_CX					8
#define BG_CY					8
#define BG_SCREEN_CX			30
#define BG_SCREEN_CY			20
#define BG_WORLD_SCREEN_CX		32
#define BG_WORLD_SCREEN_CY		32
#define BG_WORLD_SCREEN_WCX		256
#define BG_WORLD_SCREEN_WCY		256

//is
#define STATE_INIT 				1
#define STATE_READY				2
#define STATE_GO   				3
#define STATE_INPUT				4
#define STATE_END  				5

//Panel
#define PANEL_UP_SX				14
#define PANEL_UP_SY				 4
#define PANEL_DOWN_SX			14
#define PANEL_DOWN_SY			12
#define PANEL_LEFT_SX			10
#define PANEL_LEFT_SY			 8
#define PANEL_RIGHT_SX			18
#define PANEL_RIGHT_SY			 8

#define TURN_MAX_CNT_SX			0
#define TURN_MAX_CNT_SY			19

//Panel BG
#define BG_TILE_BLACK			0
#define BG_TILE_GREEN			2
#define BG_TILE_BLUE			4
#define BG_TILE_YELLOW			6
#define BG_TILE_PINK			8
#define BG_TILE_BROWN			10
#define BG_TILE_CLEAR			12

//̓L[̃EFCg
#define KEY_REPEAT_TIME			500
#define TIMER_INTMS				10

#define TURN_MAX_CNT			50
#define KEY_MAX_CNT				4
#define SRAM_TURN_OFFSET		4
#define SRAM_RANDOM_OFFSET		0

//---------------------------------------------------------------------------
typedef struct {
	u8 Chr;
	u8 Map;
	u8 Col;
} ST_BG;

typedef struct {
	vu32 clockTick;
	vu32 wait;
	vu32 repeat;
} ST_TIMER;

typedef struct {
	u8 keyNow;							//݂̓
	u8 keyOld;							//1O̓
} ST_INPUT;

typedef struct {
	u8   key[TURN_MAX_CNT];
	s8   keyCnt;
	s8   keyInputCnt;
	bool isOver;
} ST_TURN;

//---------------------------------------------------------------------------
ST_BG    BG2;
ST_BG    BG3;
ST_TIMER Timer;
ST_INPUT Input;
ST_TURN  Trun;

u8  State;								//istO
u32 Random;
//---------------------------------------------------------------------------
void StateInit();
void StateReady();
void StateGo();
void StateInput();
void StateEnd();

void InitWS();
void InitBG();
void InitTurn();
void InitRandom();
void InitTimer();
void IRQUserHandler();

void DrawFrame();
void DrawPanel(u8 key, bool isBrown);
void DrawSramTurn();

u8   GetRandomKey();

void ProcWaitDrawPanel(u8 key, u32 wait1, u32 wait2);
bool ProcKeyInput();

void WaitMs(u32 wait);

bool WaitTimer();
bool RepeatKey();

void SramSaveTurn();
u8   SramLoadTurn();
bool SramCompTurn();
void SramSaveRandom();
u32  SramLoadRandom();

//---------------------------------------------------------------------------
int main(void)
{
	State = STATE_INIT;

	for(;;)
	{
		WaitForVsync();

		switch(State)
		{
			case STATE_INIT:  StateInit();  break;
			case STATE_READY: StateReady(); break;
			case STATE_GO:    StateGo();    break;
			case STATE_INPUT: StateInput(); break;
			case STATE_END:   StateEnd();   break;
			default:
				DebugMessage((u8*)"Err: Main State");
				break;
		}
	}

	return 0;
}
//---------------------------------------------------------------------------
void StateInit()
{
	KeyInit();

	InitWS();
	InitBG();
	InitTimer();
	InitTurn();
	InitRandom();

	DrawFrame();
	DrawSramTurn();

	SetMode(MODE_0 | BG2_ENABLE | BG3_ENABLE | OBJ_ENABLE | OBJ_MAP_2D);
	State = STATE_READY;
}
//---------------------------------------------------------------------------
//EFCgRg[
void InitWS()
{
	REG_WSCNT = CST_ROM0_1ST_3WAIT | CST_ROM0_2ND_1WAIT;
}
//---------------------------------------------------------------------------
void InitBG()
{
	//BG2 ^C
	//BG3 \

	BG2.Chr = 1;
	BG2.Map = 0;
	BG2.Col = 1;

	BG3.Chr = 2;
	BG3.Map = 1;
	BG3.Col = 0;

	REG_BG2CNT = (BG_SIZEA_256_256     | BG_COLOR_16         |
				  BG_CHARBASE(BG2.Chr) | BG_MAPBASE(BG2.Map) | 0);

	REG_BG3CNT = (BG_SIZEA_256_256     | BG_COLOR_16         |
				  BG_CHARBASE(BG3.Chr) | BG_MAPBASE(BG3.Map) | 0);

	BGSetPal16  (BG2.Col, (u16*)&game_bg_pal_bin);
	BGSetChrData(BG2.Chr, (u16*)&game_bg_img_bin, BG_IMG_SIZE);

	BGSetPal16  (BG3.Col, (u16*)&BGFont_palette);
	BGSetChrData(BG3.Chr, (u16*)&BGFont_data, BGFont_DATA_SIZE);
}
//---------------------------------------------------------------------------
void InitTimer()
{
	Timer.wait      = 0;
	Timer.repeat    = 0;
	Timer.clockTick = 0;

	REG_INTERUPT = (u32)IRQUserHandler;
	REG_IE       = IRQ_BIT_TIMER0;
	REG_IME      = IRQ_MASTER_ON;

	REG_TM0D     = 65536 - 2621; // (((16*1024*1024) / 64) * 10) / 1000
	REG_TM0CNT   = TM_ENABLE | TM_FREQ_PER_64 | TM_USEIRQ;
}
//---------------------------------------------------------------------------
void IRQUserHandler()
{
	Timer.clockTick += TIMER_INTMS;
	REG_IF = IRQ_BIT_TIMER0;
}
//---------------------------------------------------------------------------
void InitTurn()
{
	ST_TURN* p = &Trun;
	s8 i;

	for(i=0; i<TURN_MAX_CNT; i++)
	{
		p->key[i] = 0;
	}
	p->keyCnt      = 0;
	p->keyInputCnt = 0;
	p->isOver      = FALSE;
}
//---------------------------------------------------------------------------
//L[s[hi`^Oh~j
bool RepeatKey()
{
	ST_INPUT* p = &Input;
	ST_TIMER* t = &Timer;

	if(p->keyNow && p->keyNow == p->keyOld)
	{
		if(t->clockTick - t->repeat < KEY_REPEAT_TIME)
		{
			return TRUE;
		}

		return FALSE;
	}

	p->keyOld = p->keyNow;
	t->repeat = t->clockTick;

	return FALSE;
}
//---------------------------------------------------------------------------
bool WaitTimer()
{
	if(Timer.clockTick - Timer.wait < TIMER_INTMS)
	{
		return FALSE;
	}

	Timer.wait = Timer.clockTick;
	return TRUE;
}
//---------------------------------------------------------------------------
void DrawFrame()
{
	BGDrawCls(BG3.Map, ' ');
	BGDrawCls(BG2.Map, BG_TILE_CLEAR);

	DrawPanel(KEY_UP,    TRUE);
	DrawPanel(KEY_DOWN,  TRUE);
	DrawPanel(KEY_LEFT,  TRUE);
	DrawPanel(KEY_RIGHT, TRUE);
}
//---------------------------------------------------------------------------
void DrawPanel(u8 key, bool isBrown)
{
	u8 x = 0;
	u8 y = 0;
	u8 t = 0;	//type

	switch(key)
	{
		case KEY_UP:
			x = PANEL_UP_SX;
			y = PANEL_UP_SY;
			t = BG_TILE_GREEN;
			break;

		case KEY_DOWN:
			x = PANEL_DOWN_SX;
			y = PANEL_DOWN_SY;
			t = BG_TILE_BLUE;
			break;

		case KEY_LEFT:
			x = PANEL_LEFT_SX;
			y = PANEL_LEFT_SY;
			t = BG_TILE_YELLOW;
			break;

		case KEY_RIGHT:
			x = PANEL_RIGHT_SX;
			y = PANEL_RIGHT_SY;
			t = BG_TILE_PINK;
			break;

		default:
			DebugMessage((u8*)"Err: DrawPanel type");
			break;
	}

	if(isBrown == TRUE)
	{
		t = BG_TILE_BROWN;
	}

	BGDrawChar(BG2.Map, x+0, y+0, BG2.Col, t+0 );
	BGDrawChar(BG2.Map, x+1, y+0, BG2.Col, t+1 );
	BGDrawChar(BG2.Map, x+0, y+1, BG2.Col, t+32);
	BGDrawChar(BG2.Map, x+1, y+1, BG2.Col, t+33);
}
//---------------------------------------------------------------------------
void StateReady()
{
	ST_TURN* p = &Trun;

	p->key[p->keyCnt] = GetRandomKey();
	p->keyCnt++;

	BGDrawStr(BG3.Map, 12, 16, (u8*)"Ready?");
	WaitMs(1000);

	BGDrawStr(BG3.Map, 12, 16, (u8*)"  Go! ");
	State = STATE_GO;
}
//---------------------------------------------------------------------------
u8 GetRandomKey()
{
	u8 rnd = rand() % KEY_MAX_CNT;
	u8 key = 0;

	switch(rnd)
	{
		case 0: key = KEY_UP;    break;
		case 1: key = KEY_DOWN;  break;
		case 2: key = KEY_LEFT;  break;
		case 3: key = KEY_RIGHT; break;
		default:
			DebugMessage((u8*)"Err: GetRandomKey rnd");
			break;
	}

	return key;
}
//---------------------------------------------------------------------------
void StateGo()
{
	ST_TURN* p = &Trun;

	s16 i;
	for(i=0; i<p->keyCnt; i++)
	{
		ProcWaitDrawPanel(p->key[i], 320, 128);
	}

	BGDrawStr(BG3.Map, 12, 16, (u8*)"INPUT!");
	State = STATE_INPUT;
}
//---------------------------------------------------------------------------
void StateInput()
{
	if( ProcKeyInput() == TRUE )
	{
		ST_TURN*  t = &Trun;
		ST_INPUT* p = &Input;

		ProcWaitDrawPanel(p->keyNow, 160, 0);

		//̓~X̃`FbN܂
		if(p->keyNow != t->key[t->keyInputCnt])
		{
			t->isOver = TRUE;
			State = STATE_END;
			return;
		}


		//o肵肪ׂăNA`FbN܂
		t->keyInputCnt++;
		if(t->keyInputCnt >= t->keyCnt)
		{
			State = STATE_END;
		}
	}

	Random = rand();	//Q[Jnɗ𔭐
}
//---------------------------------------------------------------------------
//͂m肵ꍇATRUEԂ܂
bool ProcKeyInput()
{
	//^C}[EFCg
	if( WaitTimer() == FALSE )
	{
		return FALSE;
	}

	ST_INPUT* p = &Input;
	p->keyNow   = KeyInput();

	//s[h
	if( RepeatKey() == FALSE )
	{
		return FALSE;
	}

	//̃{^Ȃ悤ɂ܂
	if(p->keyNow & KEY_UP   ) p->keyNow = KEY_UP;
	if(p->keyNow & KEY_DOWN ) p->keyNow = KEY_DOWN;
	if(p->keyNow & KEY_LEFT ) p->keyNow = KEY_LEFT;
	if(p->keyNow & KEY_RIGHT) p->keyNow = KEY_RIGHT;

	return TRUE;
}
//---------------------------------------------------------------------------
void StateEnd()
{
	ST_TURN* t = &Trun;

	//Q[I[o`FbN܂
	if(t->isOver == FALSE)
	{
		//̖o肵܂
		t->keyInputCnt = 0;
		State = STATE_READY;
		return;
	}

	if( SramCompTurn() == TRUE )
	{
		SramSaveTurn();
		DrawSramTurn();
	}

	SramSaveRandom();	//Q[Jnɗ𔭐

	BGDrawStr(BG3.Map, 12, 16, (u8*)"  End! ");
	WaitKey(KEY_A);

	State = STATE_INIT;	//Ȃ܂łǁEEE
}
//---------------------------------------------------------------------------
void WaitMs(u32 wait)
{
	u32 t = Timer.clockTick;

	while(Timer.clockTick - t < wait)
	{
	}
}
//---------------------------------------------------------------------------
void ProcWaitDrawPanel(u8 key, u32 wait1, u32 wait2)
{
		DrawPanel(key, FALSE);
		WaitMs(wait1);

		DrawPanel(key, TRUE);
		WaitMs(wait2);
}
//---------------------------------------------------------------------------
void SramSaveTurn()
{
	SRAMWrite8(SRAM_TURN_OFFSET, Trun.keyCnt - 1);
}
//---------------------------------------------------------------------------
u8 SramLoadTurn()
{
	return SRAMRead8(SRAM_TURN_OFFSET);
}
//---------------------------------------------------------------------------
bool SramCompTurn()
{
	if( Trun.keyCnt > SRAMRead8(SRAM_TURN_OFFSET) )
	{
		return TRUE;
	}
	return FALSE;
}
//---------------------------------------------------------------------------
void DrawSramTurn()
{
	char msg[32];
	bprintf(msg, "MaxTurn:%2d", SramLoadTurn() );
	BGDrawStr(BG3.Map, TURN_MAX_CNT_SX, TURN_MAX_CNT_SY, (u8*)msg);
}
//---------------------------------------------------------------------------
void SramSaveRandom()
{
	SRAMWrite32(SRAM_RANDOM_OFFSET, Random);
}
//---------------------------------------------------------------------------
u32 SramLoadRandom()
{
	return SRAMRead8(SRAM_RANDOM_OFFSET);
}
//---------------------------------------------------------------------------
void InitRandom()
{
	u32 r = SramLoadRandom();
	srand(r);
}
